/*
 Copyright (C) 2006 Grame

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 Grame Research Laboratory, 9 rue du Garet, 69001 Lyon - France
 grame@grame.fr
*/

#ifndef __JackMutex__
#define __JackMutex__

#ifdef WIN32
#include <windows.h>
#else
#include <pthread.h>
#endif

#include <assert.h>
#include "JackError.h"

namespace Jack
{

class JackMutex
{

    private:

#ifdef WIN32
        HANDLE fMutex;
#else
        pthread_mutex_t fMutex;
#endif

    public:

#ifdef WIN32

        JackMutex()
        {
            // In recursive mode by default
            fMutex = CreateMutex(0, FALSE, 0);
        }
        ~JackMutex()
        {
            CloseHandle(fMutex);
        }

        void Lock()
        {
            DWORD dwWaitResult = WaitForSingleObject(fMutex, INFINITE);
        }

        bool Trylock()
        {
           return (WAIT_OBJECT_0 == WaitForSingleObject(fMutex, 0));
        }

        void Unlock()
        {
            ReleaseMutex(fMutex);
        }

#else

        JackMutex()
        {
            // Use recursive mutex
            pthread_mutexattr_t mutex_attr;
            int res;
            res = pthread_mutexattr_init(&mutex_attr);
            assert(res == 0);
            res = pthread_mutexattr_settype(&mutex_attr, PTHREAD_MUTEX_RECURSIVE);
            assert(res == 0);
            res = pthread_mutex_init(&fMutex, &mutex_attr);
            assert(res == 0);
            res = pthread_mutexattr_destroy(&mutex_attr);
            assert(res == 0);
        }
        ~JackMutex()
        {
            pthread_mutex_destroy(&fMutex);
        }

        void Lock()
        {
            int res = pthread_mutex_lock(&fMutex);
            if (res != 0)
                jack_error("JackMutex::Lock res = %d", res);
        }

        bool Trylock()
        {
            return (pthread_mutex_trylock(&fMutex) == 0);
        }

        void Unlock()
        {
            int res = pthread_mutex_unlock(&fMutex);
            if (res != 0)
                jack_error("JackMutex::Unlock res = %d", res);
        }

#endif
};

class JackLockAble
{

    private:

        JackMutex fMutex;

    protected:

        JackLockAble()
        {}
        ~JackLockAble()
        {}

    public:
     
        void Lock()
        {
            fMutex.Lock();
        }

        bool Trylock()
        {
            return fMutex.Trylock();
        }

        void Unlock()
        {
            fMutex.Unlock();
        }

};

class JackLock
{
    private:

        JackLockAble* fObj;

    public:

        JackLock(JackLockAble* obj): fObj(obj)
        {
            fObj->Lock();
        }

        JackLock(const JackLockAble* obj): fObj((JackLockAble*)obj)
        {
            fObj->Lock();
        }

        ~JackLock()
        {
            fObj->Unlock();
        }
};


} // namespace

#endif
